/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id: $ */

package org.apache.fop.fonts;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class ArabicScriptProcessor extends ScriptProcessor
{

    ArabicScriptProcessor ( String script ) {
	super ( script );
    }

    public GlyphSequence substitute ( GlyphSequence gs, String script, String language, Map/*<LookupSpec,GlyphSubtable[]>*/ lookups )
    {
	// finals
	gs = subFina ( gs, script, language, (GlyphSubtable[]) lookups.get ( new GlyphTable.LookupSpec ( script, language, "fina" ) ) );

	// medials
	gs = subMedi ( gs, script, language, (GlyphSubtable[]) lookups.get ( new GlyphTable.LookupSpec ( script, language, "medi" ) ) );

	// initials
	gs = subInit ( gs, script, language, (GlyphSubtable[]) lookups.get ( new GlyphTable.LookupSpec ( script, language, "init" ) ) );

	// isolates
	gs = subIsol ( gs, script, language, (GlyphSubtable[]) lookups.get ( new GlyphTable.LookupSpec ( script, language, "isol" ) ) );

	// ligatures
	gs = subLiga ( gs, script, language, (GlyphSubtable[]) lookups.get ( new GlyphTable.LookupSpec ( script, language, "liga" ) ) );

	return gs;
    }

    public int[] position ( GlyphSequence gs, String script, String language, Map/*<LookupSpec,GlyphSubtable[]>*/ lookups )
    {
	return null;
    }


    private GlyphSequence substitute ( GlyphSequence gs, String script, String language, String feature, GlyphSubtable[] sta, GlyphContextTester tester, boolean reverse )
    {
	if ( ( sta != null ) && ( sta.length > 0 ) ) {
	    CharSequence ga = gs.getGlyphs();
	    GlyphSequence.CharAssociation[] aa = gs.getAssociations();
	    // enforce subtable type constraints
	    for ( int i = 0, n = sta.length; i < n; i++ ) {
		GlyphSubtable st = sta [ i ];
		if ( ! ( st instanceof GlyphSubstitutionSubtable ) )
		    throw new IncompatibleSubtableException ( "'" + feature + "' feature requires glyph substitution subtable" );
		else if ( ! ( st instanceof GlyphSubstitutionTable.SimpleSubtable ) )
		    throw new IncompatibleSubtableException ( "'" + feature + "' feature requires simple glyph substitution subtable" );
	    }
	    // perform subtitutions using compliant subtables
	    List gsl = new ArrayList();
	    List cal = new ArrayList();
	    for ( int i = 0, n = ga.length(); i < n; i++ ) {
		int k = reverse ? ( n - i - 1 ) : i;
		GlyphSequence.CharAssociation a = aa [ k ];
		GlyphSequence iss = gs.getGlyphSubsequence ( k, k + 1 );
		GlyphSequence oss;
		if ( tester.test ( iss, a ) ) {
		    oss = doSubstitutions ( iss, script, language, sta );
		} else {
		    oss = iss;
		}
		gsl.add ( oss );
		cal.add ( a );
	    }
	    gs = new GlyphSequence ( gs.getCharacters(), gsl, cal, reverse );
	}
	return gs;
    }

    private static GlyphContextTester finalContextTester =
	new GlyphContextTester() { public boolean test ( GlyphSequence gs, GlyphSequence.CharAssociation ca ) { return inFinalContext ( gs, ca ); } };

    private GlyphSequence subFina ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	return substitute ( gs, script, language, "fina", sta, finalContextTester, true );
    }

    private static GlyphContextTester medialContextTester =
	new GlyphContextTester() { public boolean test ( GlyphSequence gs, GlyphSequence.CharAssociation ca ) { return inMedialContext ( gs, ca ); } };

    private GlyphSequence subMedi ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	return substitute ( gs, script, language, "medi", sta, medialContextTester, true );
    }
    
    private static GlyphContextTester initialContextTester =
	new GlyphContextTester() { public boolean test ( GlyphSequence gs, GlyphSequence.CharAssociation ca ) { return inInitialContext ( gs, ca ); } };

    private GlyphSequence subInit ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	return substitute ( gs, script, language, "init", sta, initialContextTester, true );
    }
    
    private static GlyphContextTester isolateContextTester =
	new GlyphContextTester() { public boolean test ( GlyphSequence gs, GlyphSequence.CharAssociation ca ) { return inIsolateContext ( gs, ca ); } };

    private GlyphSequence subIsol ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	return substitute ( gs, script, language, "isol", sta, isolateContextTester, true );
    }
    
    private GlyphSequence subLiga ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	return gs;
    }

    private static boolean inFinalContext ( GlyphSequence gs, GlyphSequence.CharAssociation a )
    {
	CharSequence cs = gs.getCharacters();
	int n;
	//System.err.println ( "+inFinalContext: [" + a.getStart() + "," + a.getEnd() + "]: " + GlyphUtils.toString ( (CharSequence) gs ) );
	if ( ( n = cs.length() ) == 0 )
	    return false;
	else if ( ( a.getStart() > 0 ) && ( a.getEnd() == n ) )
	    return true;
	else
	    return false;
    }

    private static boolean inMedialContext ( GlyphSequence gs, GlyphSequence.CharAssociation a )
    {
	CharSequence cs = gs.getCharacters();
	int n;
	//System.err.println ( "+inMedialContext: [" + a.getStart() + "," + a.getEnd() + "]: " + GlyphUtils.toString ( (CharSequence) gs ) );
	if ( ( n = cs.length() ) == 0 )
	    return false;
	else if ( ( a.getStart() > 0 ) && ( a.getEnd() < n ) )
	    return true;
	else
	    return false;
    }

    private static boolean inInitialContext ( GlyphSequence gs, GlyphSequence.CharAssociation a )
    {
	CharSequence cs = gs.getCharacters();
	int n;
	//System.err.println ( "+inInitialContext: [" + a.getStart() + "," + a.getEnd() + "]: " + GlyphUtils.toString ( (CharSequence) gs ) );
	if ( ( n = cs.length() ) == 0 )
	    return false;
	else if ( ( a.getStart() == 0 ) && ( a.getEnd() < n ) )
	    return true;
	else
	    return false;
    }

    private static boolean inIsolateContext ( GlyphSequence gs, GlyphSequence.CharAssociation a )
    {
	CharSequence cs = gs.getCharacters();
	int n;
	//System.err.println ( "+inIsolateContext: [" + a.getStart() + "," + a.getEnd() + "]: " + GlyphUtils.toString ( (CharSequence) gs ) );
	if ( ( n = cs.length() ) == 0 )
	    return false;
	else if ( ( a.getStart() == 0 ) && ( a.getEnd() == n ) )
	    return true;
	else
	    return false;
    }

    private GlyphSequence doSubstitutions ( GlyphSequence gs, String script, String language, GlyphSubtable[] sta )
    {
	//System.err.println ( "+doSubstitutions: " + GlyphUtils.toString ( (CharSequence) gs ) );
	for ( int i = 0, n = sta.length; i < n; i++ ) {
	    GlyphSubtable st = sta [ i ];
	    assert st instanceof GlyphSubstitutionSubtable;
	    gs = ( (GlyphSubstitutionSubtable) st ) . substitute ( gs, script, language );
	}
	//System.err.println ( "-doSubstitutions: " + GlyphUtils.toString ( (CharSequence) gs ) );
	return gs;
    }

}
