/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id: $ */

package org.apache.fop.fonts;

import java.nio.CharBuffer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class GlyphSubstitutionTable extends GlyphTable implements GlyphSubstitution {

    public static final int GSUB_LOOKUP_TYPE_SINGLE = 1;
    public static final int GSUB_LOOKUP_TYPE_MULTIPLE = 2;
    public static final int GSUB_LOOKUP_TYPE_ALTERNATE = 3;
    public static final int GSUB_LOOKUP_TYPE_LIGATURE = 4;
    public static final int GSUB_LOOKUP_TYPE_CONTEXT = 5;
    public static final int GSUB_LOOKUP_TYPE_CHAINING_CONTEXT = 6;
    public static final int GSUB_LOOKUP_TYPE_EXTENSION_SUBSTITUTION = 7;
    public static final int GSUB_LOOKUP_TYPE_REVERSE_CHAINING_CONTEXT_SINGLE = 8;

    public GlyphSubstitutionTable ( Map lookups, List subtables ) {
	super ( lookups );
	if ( ( subtables == null ) || ( subtables.size() == 0 ) )
	    throw new IllegalArgumentException ( "subtables must be non-empty" );
	else {
	    for ( Iterator it = subtables.iterator(); it.hasNext(); ) {
		Object o = it.next();
		if ( o instanceof GlyphSubstitutionSubtable ) {
		    addSubtable ( (GlyphSubtable) o );
		} else {
		    throw new IllegalArgumentException ( "subtable must be a glyph substitution subtable" );
		}
	    }
	}
    }

    /** {@inheritDoc} */
    public GlyphSequence substitute ( GlyphSequence gs, String script, String language )
    {
	GlyphSequence ogs;
	Map/*<LookupSpec,GlyphSubtable[]>*/ lookups = matchLookups ( script, language, "*" );
	if ( ( lookups != null ) && ( lookups.size() > 0 ) ) {
	    ScriptProcessor sp = ScriptProcessor.getInstance ( script );
	    ogs = sp.substitute ( gs, script, language, lookups );
	} else
	    ogs = gs;
	return ogs;
    }

    public static int getLookupTypeFromName ( String name ) {
	int t;
	String s = name.toLowerCase();
	if ( "single".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_SINGLE;
	else if ( "multiple".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_MULTIPLE;
	else if ( "alternate".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_ALTERNATE;
	else if ( "ligature".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_LIGATURE;
	else if ( "context".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_CONTEXT;
	else if ( "chainingcontext".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_CHAINING_CONTEXT;
	else if ( "extensionsubstitution".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_EXTENSION_SUBSTITUTION;
	else if ( "reversechainiingcontextsingle".equals ( s ) )
	    t = GSUB_LOOKUP_TYPE_REVERSE_CHAINING_CONTEXT_SINGLE;
	else
	    t = -1;
	return t;
    }

    public static String getLookupTypeName ( int type ) {
	String tn = null;
	switch ( type ) {
	case GSUB_LOOKUP_TYPE_SINGLE:
	    tn = "single";
	    break;
	case GSUB_LOOKUP_TYPE_MULTIPLE:
	    tn = "multiple";
	    break;
	case GSUB_LOOKUP_TYPE_ALTERNATE:
	    tn = "alternate";
	    break;
	case GSUB_LOOKUP_TYPE_LIGATURE:
	    tn = "ligature";
	    break;
	case GSUB_LOOKUP_TYPE_CONTEXT:
	    tn = "context";
	    break;
	case GSUB_LOOKUP_TYPE_CHAINING_CONTEXT:
	    tn = "chainingcontext";
	    break;
	case GSUB_LOOKUP_TYPE_EXTENSION_SUBSTITUTION:
	    tn = "extensionsubstitution";
	    break;
	case GSUB_LOOKUP_TYPE_REVERSE_CHAINING_CONTEXT_SINGLE:
	    tn = "reversechainiingcontextsingle";
	    break;
	default:
	    tn = "unknown";
	    break;
	}
	return tn;
    }

    public static GlyphSubtable createSubtable ( int type, String id, int sequence, int flags, int format, List coverage, List entries )
    {
	GlyphSubtable st = null;
	switch ( type ) {
	case GSUB_LOOKUP_TYPE_SINGLE:
	    st = new SimpleSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_MULTIPLE:
	    st = new MultipleSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_ALTERNATE:
	    st = new AlternateSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_LIGATURE:
	    st = new LigatureSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_CONTEXT:
	    st = new ContextSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_CHAINING_CONTEXT:
	    st = new ChainingContextSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_EXTENSION_SUBSTITUTION:
	    st = new ExtensionSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	case GSUB_LOOKUP_TYPE_REVERSE_CHAINING_CONTEXT_SINGLE:
	    st = new ReverseChainingSingleSubtable ( id, sequence, flags, format, coverage, entries );
	    break;
	default:
	    break;
	}
	return st;
    }

    static class SimpleSubtable extends GlyphSubstitutionSubtable {
	private int[] map;
	public SimpleSubtable ( String id, int sequence, int flags, int format, List coverage, List entries ) {
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	    populate ( entries );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_SINGLE;
	}
	public List getEntries() {
	    List entries = new ArrayList ( map.length );
	    for ( int i = 0, n = map.length; i < n; i++ )
		entries.add ( new Integer ( map[i] ) );
	    return entries;
	}
	public GlyphSequence substitute ( GlyphSequence gs, String script, String language ) {
	    CharBuffer cb = CharBuffer.allocate ( gs.length() );
	    for ( int i = 0; i < gs.length(); i++ ) {
		int gi = gs.charAt ( i );
		int ci, go = gi;
		if ( ( ci = getCoverageIndex ( gi ) ) >= 0 ) {
		    assert ci < map.length : "coverage index out of range";
		    if ( ci < map.length )
			go = map [ ci ];
		}
		if ( ( go < 0 ) || ( go > 65535 ) )
		    go = 65535;
		cb.put ( (char) go );
	    }
	    cb.rewind();
	    return new GlyphSequence ( gs.getCharacters(), (CharSequence) cb, null );
	}
	private void populate ( List entries ) {
	    int i = 0, n = entries.size();
	    int[] map = new int [ n ];
	    for ( Iterator it = entries.iterator(); it.hasNext(); ) {
		Object o = it.next();
		if ( o instanceof Integer ) {
		    int gid = ( (Integer) o ) . intValue();
		    if ( ( gid >= 0 ) && ( gid < 65536 ) ) {
			map [ i++ ] = gid;
		    } else {
			throw new IllegalArgumentException ( "illegal glyph index: " + gid );
		    }
		} else {
		    throw new IllegalArgumentException ( "illegal entries entry, must be Integer: " + o );
		}
	    }
	    assert i == n;
	    assert this.map == null;
	    this.map = map;
	}
	public String toString() {
	    StringBuffer sb = new StringBuffer(super.toString());
	    sb.append('{');
	    sb.append("coverage=");
	    sb.append(getCoverage().toString());
	    sb.append(",entries={");
	    for ( int i = 0, n = map.length; i < n; i++ ) {
		if ( i > 0 )
		    sb.append(',');
		sb.append(Integer.toString(map[i]));
	    }
	    sb.append('}');
	    sb.append('}');
	    return sb.toString();
	}
    }

    static class MultipleSubtable extends GlyphSubstitutionSubtable {
	public MultipleSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_MULTIPLE;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    static class AlternateSubtable extends GlyphSubstitutionSubtable {
	public AlternateSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_ALTERNATE;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    static class LigatureSubtable extends GlyphSubstitutionSubtable {
	private LigatureSet[] map;
	public LigatureSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	    populate ( entries );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_LIGATURE;
	}
	public List getEntries() {
	    List entries = new ArrayList ( map.length );
	    for ( int i = 0, n = map.length; i < n; i++ )
		entries.add ( map[i] );
	    return entries;
	}
	public GlyphSequence substitute ( GlyphSequence gs, String script, String language ) {
	    CharBuffer cb = CharBuffer.allocate ( gs.length() );
	    for ( int i = 0, n = gs.length(); i < n; i++ ) {
		int gi = gs.charAt ( i );
		int ci, go = gi;
		LigatureSet ls = null;
		if ( ( ci = getCoverageIndex ( gi ) ) >= 0 ) {
		    assert ci < map.length : "coverage index out of range";
		    if ( ci < map.length )
			ls = map [ ci ];
		}
		if ( ls != null ) {
		    Ligature l;
		    if ( ( l = findLigature ( ls, gs, i ) ) != null ) {
			go = l.getLigature();
			i += l.getNumComponents();
		    }
		}
		if ( ( go < 0 ) || ( go > 65535 ) )
		    go = 65535;
		cb.put ( (char) go );
	    }
	    cb.rewind();
	    return new GlyphSequence ( gs.getCharacters(), (CharSequence) cb, null );
	}
	private void populate ( List entries ) {
	    int i = 0, n = entries.size();
	    LigatureSet[] map = new LigatureSet [ n ];
	    for ( Iterator it = entries.iterator(); it.hasNext(); ) {
		Object o = it.next();
		if ( o instanceof LigatureSet ) {
		    map [ i++ ] = (LigatureSet) o;
		} else {
		    throw new IllegalArgumentException ( "illegal ligatures entry, must be LigatureSet: " + o );
		}
	    }
	    assert i == n;
	    assert this.map == null;
	    this.map = map;
	}
	private Ligature findLigature ( LigatureSet ls, CharSequence cs, int offset ) {
	    Ligature[] la = ls.getLigatures();
	    int k = -1;
	    int maxComponents = -1;
	    for ( int i = 0, n = la.length; i < n; i++ ) {
		Ligature l = la [ i ];
		if ( l.matchesComponents ( cs, offset + 1 ) ) {
		    int nc = l.getNumComponents();
		    if ( nc > maxComponents ) {
			maxComponents = nc;
			k = i;
		    }
		}
	    }
	    if ( k >= 0 ) {
		return la [ k ];
	    } else
		return null;
	}
	public String toString() {
	    StringBuffer sb = new StringBuffer(super.toString());
	    sb.append('{');
	    sb.append("coverage=");
	    sb.append(getCoverage().toString());
	    sb.append(",entries={");
	    for ( int i = 0, n = map.length; i < n; i++ ) {
		if ( i > 0 )
		    sb.append(',');
		sb.append(map[i]);
	    }
	    sb.append('}');
	    sb.append('}');
	    return sb.toString();
	}
    }

    static class ContextSubtable extends GlyphSubstitutionSubtable {
	public ContextSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_CONTEXT;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    static class ChainingContextSubtable extends GlyphSubstitutionSubtable {
	public ChainingContextSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_CHAINING_CONTEXT;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    static class ExtensionSubtable extends GlyphSubstitutionSubtable {
	public ExtensionSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_EXTENSION_SUBSTITUTION;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    static class ReverseChainingSingleSubtable extends GlyphSubstitutionSubtable {
	public ReverseChainingSingleSubtable ( String id, int sequence, int flags, int format, List coverage, List entries )
	{
	    super ( id, sequence, flags, format, GlyphCoverageTable.createCoverageTable ( coverage ) );
	}
	public int getType() {
	    return GSUB_LOOKUP_TYPE_REVERSE_CHAINING_CONTEXT_SINGLE;
	}
	public List getEntries() {
	    return null; // [TBD] - implement me
	}
    }

    public static class Ligature {

	private final int ligature;			// (resulting) ligature glyph 
	private final int[] components;			// component glyph codes (note that first component is implied)

	public Ligature ( int ligature, int[] components ) {
	    if ( ( ligature < 0 ) || ( ligature > 65535 ) )
		throw new IllegalArgumentException ( "invalid ligature glyph index: " + ligature );
	    else if ( ( components == null ) || ( components.length == 0 ) )
		throw new IllegalArgumentException ( "invalid ligature components, must be non-empty array" );
	    else {
		for ( int i = 0, n = components.length; i < n; i++ ) {
		    int c = components [ i ];
		    if ( ( c < 0 ) || ( c > 65535 ) )
			throw new IllegalArgumentException ( "invalid component glyph index: " + c );
		}
		this.ligature = ligature;
		this.components = components;
	    }
	}

	public int getLigature() {
	    return ligature;
	}

	public int[] getComponents() {
	    return components;
	}

	public int getNumComponents() {
	    return components.length;
	}

	public boolean matchesComponents ( CharSequence cs, int offset ) {
	    if ( ( offset + components.length ) > cs.length() )
		return false;
	    else {
		for ( int i = 0, n = components.length; i < n; i++ ) {
		    if ( (int) cs.charAt ( offset + i ) != components [ i ] )
			return false;
		}
		return true;
	    }
	}

	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append("{components={");
	    for ( int i = 0, n = components.length; i < n; i++ ) {
		if ( i > 0 )
		    sb.append(',');
		sb.append(Integer.toString(components[i]));
	    }
	    sb.append("},ligature=");
	    sb.append(Integer.toString(ligature));
	    sb.append("}");
	    return sb.toString();
	}

    }

    public static class LigatureSet {

	private final Ligature[] ligatures;			// set of ligatures all of which share the first (implied) component

	public LigatureSet ( List ligatures ) {
	    this ( (Ligature[]) ligatures.toArray ( new Ligature [ 0 ] ) );
	}

	public LigatureSet ( Ligature[] ligatures ) {
	    if ( ( ligatures == null ) || ( ligatures.length == 0 ) )
		throw new IllegalArgumentException ( "invalid ligatures, must be non-empty array" );
	    else {
		this.ligatures = ligatures;
	    }
	}

	public Ligature[] getLigatures() {
	    return ligatures;
	}

	public int getNumLigatures() {
	    return ligatures.length;
	}

	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append("{ligs={");
	    for ( int i = 0, n = ligatures.length; i < n; i++ ) {
		if ( i > 0 )
		    sb.append(',');
		sb.append(ligatures[i]);
	    }
	    sb.append("}}");
	    return sb.toString();
	}

    }


}
